from openfisca_uk.model_api import *


class DLA(Variable):
    value_type = float
    entity = Person
    label = "Disability Living Allowance"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return add(person, period, ["DLA_M", "DLA_SC"])


class DLA_M(Variable):
    value_type = float
    entity = Person
    label = "Disability Living Allowance (mobility component)"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("DLA_M_reported", period)


class DLA_SC(Variable):
    value_type = float
    entity = Person
    label = "Disability Living Allowance (self-care)"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("DLA_SC_reported", period)


class DLA_SC_middle_plus(Variable):
    value_type = bool
    entity = Person
    label = "Receives at least DLA (self-care) middle rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        DLA_SC = parameters(period).benefit.DLA.self_care
        return person("DLA_SC", period) >= DLA_SC.middle


class DLA_M_reported(Variable):
    value_type = float
    entity = Person
    label = "Disability Living Allowance (mobility component) (reported)"
    definition_period = YEAR
    unit = "currency-GBP"


class DLA_SC_reported(Variable):
    value_type = float
    entity = Person
    label = "Disability Living Allowance (self-care) (reported)"
    definition_period = YEAR
    unit = "currency-GBP"



from openfisca_uk_data.datasets.frs.frs_enhanced.frs_enhanced import (
    FRSEnhanced,
)
from openfisca_uk.model_api import *


class baseline_hbai_excluded_income(Variable):
    label = "HBAI-excluded income (baseline)"
    documentation = "Total value of income not included in HBAI household net income in the baseline"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        # Establish if currently running a microsimulation
        if len(household.nb_persons()) > 1_000:
            from openfisca_uk import Microsimulation

            # Simulate baseline policy
            result = Microsimulation(
                dataset=FRSEnhanced, year=2019
            ).simulation.calculate("hbai_excluded_income", period)
            # Check that the dataset/year combination is valid
            # (i.e. that the arrays are the same size)
            if len(result) == len(household.nb_persons()):
                return result
        # If baseline policy not viable from the above method,
        # no change in HBAI excluded income
        return household("hbai_excluded_income", period)


class hbai_excluded_income(Variable):
    label = "HBAI-excluded income"
    documentation = (
        "Total value of income not included in HBAI household net income"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        VARIABLES = [
            "expected_sdlt",
            "expected_ltt",
            "expected_lbtt",
            "business_rates",
        ]
        return -add(household, period, VARIABLES)


class hbai_excluded_income_change(Variable):
    label = "Change in HBAI-excluded income"
    documentation = "Effect of policy reforms on HBAI-excluded income"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        hbai_excluded_income = household("hbai_excluded_income", period)
        baseline_hbai_excluded_income = household(
            "baseline_hbai_excluded_income", period
        )
        return hbai_excluded_income - baseline_hbai_excluded_income



from openfisca_uk.model_api import *


class NI_class_4(Variable):
    value_type = float
    entity = Person
    label = "Class 4 Contributions for National Insurance for the year"
    definition_period = YEAR
    reference = "Social Security and Benefits Act 1992 s. 15"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        class_4 = parameters(period).tax.national_insurance.class_4
        self_employment_income = person("self_employment_income", period)
        employee_NI = person("employee_NI", period)
        profits = self_employment_income - employee_NI
        main_amount = amount_between(
            profits,
            class_4.thresholds.lower_profits_limit,
            class_4.thresholds.upper_profits_limit,
        )
        add_amount = amount_over(
            profits, class_4.thresholds.upper_profits_limit
        )
        main_charge = main_amount * class_4.rates.main
        add_charge = add_amount * class_4.rates.additional
        return main_charge + add_charge


class employee_NI(Variable):
    value_type = float
    entity = Person
    label = "Employee-side National Insurance"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("employee_NI_class_1", period)


class self_employed_NI(Variable):
    value_type = float
    entity = Person
    label = "Self-employed National Insurance"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return add(person, period, ("NI_class_2", "NI_class_4"))


class national_insurance(Variable):
    value_type = float
    entity = Person
    label = "National Insurance"
    documentation = "Total National Insurance contributions"
    definition_period = YEAR
    unit = "currency-GBP"
    reference = "Social Security and Benefits Act 1992 s. 1(2)"

    def formula(person, period, parameters):
        CLASSES = ["employee_NI", "self_employed_NI"]
        total = add(person, period, CLASSES)
        return total * ~person("NI_exempt", period)



from openfisca_uk.model_api import *


class IIDB(Variable):
    value_type = float
    entity = Person
    label = "Industrial Injuries Disablement Benefit"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("IIDB_reported", period)


class IIDB_reported(Variable):
    value_type = float
    entity = Person
    label = "Industrial Injuries Disablement Benefit (reported)"
    definition_period = YEAR
    unit = "currency-GBP"



from openfisca_uk.model_api import *


class winter_fuel_allowance_reported(Variable):
    value_type = float
    entity = Person
    label = "Winter fuel allowance"
    definition_period = YEAR
    unit = "currency-GBP"


class winter_fuel_allowance(Variable):
    label = "Winter Fuel Allowance"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        return aggr(household, period, ["winter_fuel_allowance_reported"])



from openfisca_uk.model_api import *

"""
This file contains variables that are commonly used in benefit eligibility calculations.
"""


class family_benefits(Variable):
    value_type = float
    entity = Person
    label = "Total simulated family benefits for this person"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        FAMILY_BENEFITS = [
            "child_benefit",
            "ESA_income",
            "housing_benefit",
            "income_support",
            "JSA_income",
            "pension_credit",
            "universal_credit",
            "council_tax_benefit",
            "child_tax_credit",
            "working_tax_credit",
        ]
        benefits = add(person.benunit, period, FAMILY_BENEFITS)
        return benefits * person("is_benunit_head", period)


class family_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Total reported family benefits for this person"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        FAMILY_BENEFITS = [
            "child_benefit",
            "ESA_income",
            "housing_benefit",
            "income_support",
            "JSA_income",
            "pension_credit",
            "universal_credit",
            "working_tax_credit",
            "child_tax_credit",
        ]
        return add(person, period, [i + "_reported" for i in FAMILY_BENEFITS])


class benefits(Variable):
    value_type = float
    entity = Person
    label = "Total benefits"
    documentation = "Total state benefits"
    unit = "currency-GBP"
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(person, period, ["personal_benefits", "family_benefits"])


class household_benefits(Variable):
    value_type = float
    entity = Household
    label = "Benefits"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(household, period, parameters):
        HOUSEHOLD_BENEFITS = [
            "winter_fuel_allowance",
        ]
        personal_benefits = aggr(household, period, ["benefits"])
        household_benefits = add(household, period, HOUSEHOLD_BENEFITS)
        return personal_benefits + household_benefits


class other_benefits(Variable):
    value_type = float
    entity = Person
    label = "Income from benefits not modelled or detailed in the model"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        reported = person("benefits_reported", period)
        personal_family_benefits = add(
            person, period, ["personal_benefits", "family_benefits"]
        )
        return reported - personal_family_benefits


class benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Total simulated"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        BENS = ["personal_benefits_reported", "family_benefits_reported"]
        return add(person, period, BENS)


class benefits_modelling(Variable):
    value_type = float
    entity = Person
    label = (
        "Difference between reported and simulated benefits for this person"
    )
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("benefits", period) - person("benefits_reported", period)


class is_QYP(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is a qualifying young person for benefits purposes"
    definition_period = YEAR

    def formula(person, period, parameters):
        education = person("current_education", period)
        under_20 = person("age", period) < 20
        in_education = ~(
            education == education.possible_values.NOT_IN_EDUCATION
        )
        return under_20 & in_education


class is_child_or_QYP(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is a child or qualifying young person for most benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (person("age", period) < 16) | person("is_QYP", period)


class benefits_premiums(Variable):
    value_type = float
    entity = BenUnit
    label = "Value of premiums for disability and carer status"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        PREMIUMS = [
            "disability_premium",
            "enhanced_disability_premium",
            "severe_disability_premium",
            "carer_premium",
        ]
        return add(benunit, period, PREMIUMS)


class benunit_weekly_hours(Variable):
    value_type = float
    entity = BenUnit
    label = "Average weekly hours worked by adults in the benefit unit"
    definition_period = YEAR
    unit = "hour"

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["weekly_hours"])


class is_single(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether this benefit unit contains a single claimant for benefits"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.SINGLE


class is_couple(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether this benefit unit contains a joint couple claimant for benefits"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.COUPLE


class is_lone_parent(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether the family is a lone parent family"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.LONE_PARENT


class is_single_person(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether the family is a single person"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.SINGLE


class personal_benefits(Variable):
    value_type = float
    entity = Person
    label = "Value of personal, non-means-tested benefits"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        BENEFITS = [
            "AA",
            "AFCS",
            "BSP",
            "carers_allowance",
            "DLA_M",
            "DLA_SC",
            "ESA_contrib",
            "IIDB",
            "incapacity_benefit",
            "JSA_contrib",
            "PIP_M",
            "PIP_DL",
            "SDA",
            "state_pension",
            "student_payments",
            "student_loans",
            "maternity_allowance",
            "SSP",
            "SMP",
            "ssmg",
        ]
        return add(person, period, BENEFITS)


class personal_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Value of personal, non-means-tested benefits"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        BENEFITS = [
            "AA",
            "AFCS",
            "BSP",
            "carers_allowance",
            "DLA_M",
            "DLA_SC",
            "ESA_contrib",
            "IIDB",
            "incapacity_benefit",
            "JSA_contrib",
            "PIP_M",
            "PIP_DL",
            "SDA",
            "state_pension",
        ]
        return add(person, period, [i + "_reported" for i in BENEFITS])


class claims_all_entitled_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = "Claims all eligible benefits"
    definition_period = YEAR
    documentation = (
        "Whether this family would claim any benefit they are entitled to"
    )


class claims_legacy_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = "Claims legacy benefits"
    documentation = "Whether this family is currently receiving legacy benefits (overrides UC claimant status)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        # Assign legacy/UC claimant status, consistently for each household
        household = benunit.members.household
        benunit_random = benunit.value_from_first_person(
            household.project(random(household))
        )
        UC_rollout = parameters(period).benefit.universal_credit.rollout_rate
        return benunit_random > UC_rollout



from openfisca_uk.model_api import *


class ESA_income_reported(Variable):
    value_type = float
    entity = Person
    label = "ESA (income-based) (reported amount)"
    definition_period = YEAR
    unit = "currency-GBP"


class would_claim_ESA_income(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim income-based ESA"
    documentation = "Whether this family would claim income-based Employment Support Allowance if eligible"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reports_ESA_income = aggr(benunit, period, ["ESA_income_reported"]) > 0
        claims_all = benunit("claims_all_entitled_benefits", period)
        return reports_ESA_income | claims_all


class ESA_income_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "ESA (income) eligible"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["ESA_income_reported"]) > 0


class claims_ESA_income(Variable):
    value_type = bool
    entity = BenUnit
    label = "Claims ESA (income)"
    documentation = "Claims income-based Employment and Support Allowance"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        would_claim = benunit("would_claim_ESA_income", period)
        claims_legacy_benefits = benunit("claims_legacy_benefits", period)
        return would_claim & claims_legacy_benefits


class ESA_income(Variable):
    value_type = float
    entity = BenUnit
    label = "ESA (income-based)"
    documentation = "Employment and Support Allowance"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["ESA_income_reported"])



import logging
from openfisca_uk import CountryTaxBenefitSystem
from openfisca_uk.entities import entities
import numpy as np
import warnings
from openfisca_uk.entities import *
import numpy as np
import warnings
from openfisca_uk.initial_setup import set_default
from openfisca_uk.tools.parameters import backdate_parameters
from openfisca_tools import ReformType
from openfisca_uk_data import DATASETS, SynthFRS
from openfisca_tools.microsimulation import (
    Microsimulation as GeneralMicrosimulation,
)
from openfisca_tools.hypothetical import IndividualSim as GeneralIndividualSim
import yaml
from pathlib import Path


with open(Path(__file__).parent / "datasets.yml") as f:
    datasets = yaml.safe_load(f)
    DEFAULT_DATASET = list(
        filter(lambda ds: ds.name == datasets["default"], DATASETS)
    )[0]


warnings.filterwarnings("ignore")

np.random.seed(0)


class Microsimulation(GeneralMicrosimulation):
    tax_benefit_system = CountryTaxBenefitSystem
    entities = entities
    default_dataset = DEFAULT_DATASET
    post_reform = backdate_parameters()

    def __init__(
        self, reform: ReformType = (), dataset: type = None, year: int = None
    ):
        if dataset is None:
            dataset = self.default_dataset
        else:
            dataset = dataset
        if year is None:
            year = self.default_year or max(dataset.years)
        else:
            year = year

        # Check if dataset is available

        if year not in dataset.years:
            download = input(
                f"\nYear {year} not available in dataset {dataset.name}: \n\t* Download the dataset [y]\n\t* Use the synthetic FRS (and set default) [n]\n\nChoice: "
            )
            if download == "y":
                dataset.download(year)
            else:
                set_default(SynthFRS)
                dataset = SynthFRS
                if year not in dataset.years:
                    logging.info(
                        f"Year {year} synthetic FRS not stored, downloading..."
                    )
                    dataset.download(year)

        super().__init__(reform=reform, dataset=dataset, year=year)


class IndividualSim(GeneralIndividualSim):
    tax_benefit_system = CountryTaxBenefitSystem
    post_reform = backdate_parameters()



from openfisca_uk.model_api import *


class weekly_rent(Variable):
    value_type = float
    entity = Household
    label = "Weekly average rent"
    definition_period = YEAR
    unit = "currency-GBP"


@uprated(by="CPI")
class benunit_rent(Variable):
    value_type = float
    entity = BenUnit
    label = "Rent"
    documentation = "Gross rent that members of this family are liable for"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["personal_rent"])


class personal_rent(Variable):
    value_type = float
    entity = Person
    label = "Rent liable"
    documentation = "The gross rent this person is liable for"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        rent = person.household("rent", period)
        is_household_head = person("is_household_head", period)
        return rent * is_household_head


class family_rent(Variable):
    value_type = float
    entity = BenUnit
    label = "Gross rent for the family"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        personal_rent = benunit.members("personal_rent", period)
        return benunit.sum(personal_rent)


class childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = "Cost of childcare"
    definition_period = YEAR
    unit = "currency-GBP"


class private_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Private pension contributions"
    definition_period = YEAR
    unit = "currency-GBP"


class occupational_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Occupational pension contributions"
    definition_period = YEAR
    unit = "currency-GBP"


class housing_service_charges(Variable):
    value_type = float
    entity = Household
    label = "Housing service charges"
    definition_period = YEAR
    unit = "currency-GBP"


class water_and_sewerage_charges(Variable):
    value_type = float
    entity = Household
    label = "Water and sewerage charges"
    definition_period = YEAR


class employer_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Employer pension contributions"
    definition_period = YEAR
    unit = "currency-GBP"


class weekly_childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = "Average cost of childcare"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("childcare_expenses", period) / WEEKS_IN_YEAR


@uprated(by="CPI")
class housing_costs(Variable):
    value_type = float
    entity = Household
    label = "Total housing costs"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(household, period, parameters):
        return add(household, period, ["rent", "mortgage"])


class maintenance_expenses(Variable):
    value_type = float
    entity = Person
    label = "Maintenance expenses"
    definition_period = YEAR
    unit = "currency-GBP"


class mortgage_interest_repayment(Variable):
    value_type = float
    entity = Household
    label = "Total mortgage payments"
    definition_period = YEAR
    unit = "currency-GBP"


class mortgage_capital_repayment(Variable):
    value_type = float
    entity = Household
    label = "Mortgage payments"
    definition_period = YEAR
    unit = "currency-GBP"


class mortgage(Variable):
    value_type = float
    entity = Household
    label = "Total mortgage payments"
    definition_period = YEAR
    unit = "currency-GBP"


@uprated(by="council_tax")
class council_tax(Variable):
    value_type = float
    entity = Household
    label = "Council Tax"
    definition_period = YEAR
    unit = "currency-GBP"
    quantity_type = FLOW


class council_tax_less_benefit(Variable):
    label = "Council Tax (less CTB)"
    documentation = "Council Tax minus the Council Tax Benefit"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        person = household.members
        council_tax_benefit = household.sum(
            person.benunit("council_tax_benefit", period)
            * person("is_benunit_head", period)
        )
        return household("council_tax", period) - council_tax_benefit


class CouncilTaxBand(Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"
    I = "I"


class council_tax_band(Variable):
    value_type = Enum
    possible_values = CouncilTaxBand
    default_value = CouncilTaxBand.D
    entity = Household
    label = "Council Tax Band"
    definition_period = YEAR



from openfisca_uk.model_api import *

"""
This file calculates the overall liability for Income Tax.
"""


class earned_taxable_income(Variable):
    value_type = float
    entity = Person
    label = "Non-savings, non-dividend income for Income Tax"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 10"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        EXCLUSIONS = [
            "taxable_savings_interest_income",
            "taxable_dividend_income",
            "allowances",
            "marriage_allowance",
        ]
        ANI = person("adjusted_net_income", period)
        exclusions = add(person, period, EXCLUSIONS)
        return max_(0, ANI - exclusions)


class taxed_income(Variable):
    value_type = float
    entity = Person
    label = "Income which is taxed"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        COMPONENTS = [
            "earned_taxable_income",
            "taxed_savings_income",
            "taxed_dividend_income",
        ]
        return add(person, period, COMPONENTS)


class basic_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Earned income (non-savings, non-dividend) at the basic rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        return clip(income, thresholds[0], thresholds[1])


class higher_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Earned income (non-savings, non-dividend) at the higher rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        return clip(income, thresholds[1], thresholds[2]) - thresholds[1]


class add_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Earned income (non-savings, non-dividend) at the additional rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        return clip(income, thresholds[2], inf) - thresholds[2]


class basic_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on earned income at the basic rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        amount = person("basic_rate_earned_income", period)
        return parameters(period).tax.income_tax.rates.uk.rates[0] * amount


class higher_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on earned income at the higher rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        amount = person("higher_rate_earned_income", period)
        return parameters(period).tax.income_tax.rates.uk.rates[1] * amount


class add_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on earned income at the additional rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        amount = person("add_rate_earned_income", period)
        return parameters(period).tax.income_tax.rates.uk.rates[2] * amount


class earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on earned income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 10"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        return rates.uk.calc(person("earned_taxable_income", period))

    def formula_2017_04_06(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        uk_amount = rates.uk.calc(person("earned_taxable_income", period))
        scot_amount = rates.scotland.pre_starter_rate.calc(
            person("earned_taxable_income", period)
        )
        return where(scot, scot_amount, uk_amount)

    def formula_2018_04_06(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        uk_amount = rates.uk.calc(person("earned_taxable_income", period))
        scot_amount = rates.scotland.post_starter_rate.calc(
            person("earned_taxable_income", period)
        )
        return where(scot, scot_amount, uk_amount)


class TaxBand(Enum):
    NONE = "None"
    STARTER = "Starter (Scottish rates)"
    BASIC = "Basic"
    INTERMEDIATE = "Intermediate (Scottish rates)"
    HIGHER = "Higher"
    ADDITIONAL = "Additional"


class pays_scottish_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Whether the individual pays Scottish Income Tax rates"
    definition_period = YEAR

    def formula(person, period, parameters):
        country = person.household("country", period)
        countries = country.possible_values
        return person.household("country", period) == countries.SCOTLAND


class tax_band(Variable):
    value_type = Enum
    possible_values = TaxBand
    default_value = TaxBand.NONE
    entity = Person
    label = "Tax band of the individual"
    definition_period = YEAR

    def formula(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        basic = allowances + rates.uk.thresholds[0]
        higher = allowances + rates.uk.thresholds[-2]
        add = allowances + rates.uk.thresholds[-1]
        return select(
            [ANI >= add, ANI >= higher, ANI > basic, ANI <= basic],
            [TaxBand.ADDITIONAL, TaxBand.HIGHER, TaxBand.BASIC, TaxBand.NONE],
        )

    def formula_2017_04_06(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        income = ANI - allowances
        uk_band = select(
            [income < threshold for threshold in rates.uk.thresholds] + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        scottish_band = select(
            [
                income < threshold
                for threshold in rates.scotland.pre_starter_rate.thresholds
            ]
            + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        return where(scot, scottish_band, uk_band)

    def formula_2018_06_01(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        income = ANI - allowances
        uk_band = select(
            [income < threshold for threshold in rates.uk.thresholds[:3]]
            + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        scottish_band = select(
            [
                income < threshold
                for threshold in rates.scotland.post_starter_rate.thresholds[
                    :5
                ]
            ]
            + [True],
            [
                TaxBand.NONE,
                TaxBand.STARTER,
                TaxBand.BASIC,
                TaxBand.INTERMEDIATE,
                TaxBand.HIGHER,
                TaxBand.ADDITIONAL,
            ],
        )
        return where(scot, scottish_band, uk_band)


class basic_rate_savings_income_pre_starter(Variable):
    value_type = float
    entity = Person
    label = "Savings income which would otherwise be taxed at the basic rate, without the starter rate"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        savings_income_total = person(
            "taxable_savings_interest_income", period
        )
        savings_allowance = person("savings_allowance", period)
        savings_income = max_(0, savings_income_total - savings_allowance)
        other_income = person("earned_taxable_income", period)
        basic_rate_amount_with_savings = clip(
            savings_income + other_income,
            thresholds[0],
            thresholds[1],
        )
        basic_rate_amount_without_savings = clip(
            other_income, thresholds[0], thresholds[1]
        )
        return (
            basic_rate_amount_with_savings - basic_rate_amount_without_savings
        )


class savings_starter_rate_income(Variable):
    value_type = float
    entity = Person
    label = "Savings income which is tax-free under the starter rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 12"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("basic_rate_savings_income_pre_starter", period)
        limit = parameters(
            period
        ).tax.income_tax.rates.savings_starter_rate.allowance
        return max_(0, limit - income)


class basic_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = "Savings income at the basic rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        basic_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            thresholds[0],
            thresholds[1],
        )
        basic_rate_amount_without = clip(
            other_income, thresholds[0], thresholds[1]
        )
        return max_(0, basic_rate_amount_with - basic_rate_amount_without)


class higher_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = "Savings income at the higher rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        higher_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            thresholds[1],
            thresholds[2],
        )
        higher_rate_amount_without = clip(
            other_income, thresholds[1], thresholds[2]
        )
        return max_(0, higher_rate_amount_with - higher_rate_amount_without)


class add_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = "Savings income at the higher rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        thresholds = parameters(period).tax.income_tax.rates.uk.thresholds
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        add_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            thresholds[2],
            inf,
        )
        add_rate_amount_without = clip(other_income, thresholds[2], inf)
        return max_(0, add_rate_amount_with - add_rate_amount_without)


class taxed_savings_income(Variable):
    value_type = float
    entity = Person
    label = "Savings income which advances the person's income tax schedule"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        COMPONENTS = [
            "basic_rate_savings_income",
            "higher_rate_savings_income",
            "add_rate_savings_income",
        ]
        return add(person, period, COMPONENTS)


class taxed_dividend_income(Variable):
    value_type = float
    entity = Person
    label = "Dividend income which is taxed"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return max_(
            0,
            person("taxable_dividend_income", period)
            - person("dividend_allowance", period),
        )


class savings_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on savings income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates.uk.rates
        basic_rate_amount = person("basic_rate_savings_income", period)
        higher_rate_amount = person("higher_rate_savings_income", period)
        add_rate_amount = person("add_rate_savings_income", period)
        return (
            rates[0] * basic_rate_amount
            + rates[1] * higher_rate_amount
            + rates[2] * add_rate_amount
        )


class dividend_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income tax on dividend income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 13"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        other_income = person("earned_taxable_income", period) + person(
            "taxed_savings_income", period
        )
        taxable_dividends = person("taxed_dividend_income", period)
        tax_with_dividends = rates.dividends.calc(
            other_income + taxable_dividends
        )
        tax_without_dividends = rates.dividends.calc(other_income)
        return tax_with_dividends - tax_without_dividends


class income_tax_pre_charges(Variable):
    value_type = float
    entity = Person
    label = "Income Tax before any tax charges"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 23"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        COMPONENTS = [
            "earned_income_tax",
            "savings_income_tax",
            "dividend_income_tax",
        ]
        return add(person, period, COMPONENTS)


class is_higher_earner(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is the highest earner in a family"
    definition_period = YEAR

    def formula(person, period, parameters):
        income = person("adjusted_net_income", period)
        # Add noise to incomes in order to avoid ties
        return (
            person.get_rank(person.benunit, -income + random(person) * 1e-2)
            == 0
        )


class income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income Tax"
    documentation = "Total Income Tax liability"
    definition_period = YEAR
    unit = "currency-GBP"
    reference = "Income Tax Act 2007 s. 23"

    def formula(person, period, parameters):
        tax_with_hitc = add(
            person, period, ["income_tax_pre_charges", "CB_HITC"]
        )
        married_deduction = person(
            "married_couples_allowance_deduction", period
        )
        return max_(0, tax_with_hitc - married_deduction)



from typing import Union
import yaml
from pathlib import Path
from openfisca_uk import Microsimulation
from openfisca_uk_data import FRSEnhanced
import pytest

with open(Path(__file__).parent / "statistics.yaml") as f:
    statistics = yaml.load(f, Loader=yaml.SafeLoader)

sim_2018 = Microsimulation(dataset=FRSEnhanced, year=2018)
sim_2019 = Microsimulation(dataset=FRSEnhanced, year=2019)
variables = sim_2018.simulation.tax_benefit_system.variables


def get_openfisca_uk_aggregate(variable: str, year: int) -> float:
    if year == 2018:
        sim = sim_2018
    else:
        sim = sim_2019
    return sim.calc(variable, period=year).sum()


def get_openfisca_uk_caseload(variable: str, year: int) -> float:
    if year == 2018:
        sim = sim_2018
    else:
        sim = sim_2019
    return (sim.calc(variable, period=year) > 0).sum()


class StatisticTest:
    def __init__(self, variable: str, statistic: str, year: int, **kwargs):
        self.variable = variable
        self.year = year
        self.statistic = statistic
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def openfisca_uk(self):
        if self.statistic == "aggregate":
            return get_openfisca_uk_aggregate(self.variable, self.year)
        elif self.statistic == "caseload":
            return get_openfisca_uk_caseload(self.variable, self.year)
        raise ValueError(f"Unknown statistic: {self.statistic}")

    def stored_result(self, source: str):
        if self.statistic == "aggregate":
            return statistics[self.variable][source]["aggregate"][self.year]
        elif self.statistic == "caseload":
            return statistics[self.variable][source]["caseload"][self.year]
        raise ValueError(f"Unknown statistic: {self.statistic}")

    @property
    def ukmod(self):
        return self.stored_result("ukmod")

    @property
    def official(self):
        return self.stored_result("official")

    @property
    def variable_label(self):
        return variables[self.variable].label

    def test(self) -> bool:
        raise NotImplementedError("Test not specified")

    def describe(self) -> str:
        raise NotImplementedError("Description not specified")

    def __repr__(self):
        return self.describe()


class CloserThanUKMOD(StatisticTest):
    def test(self) -> bool:
        openfisca_uk_error = abs(self.openfisca_uk - self.official)
        ukmod_error = abs(self.ukmod - self.official)
        return openfisca_uk_error <= ukmod_error

    def describe(self):
        return f"OpenFisca-UK {self.variable_label} {self.statistic} at least as close to the official aggregate as UKMOD in {self.year}"


class AbsoluteErrorLessThan(StatisticTest):
    def test(self):
        openfisca_uk_error = abs(self.openfisca_uk - self.official)
        return openfisca_uk_error < self.max_error

    def describe(self):
        return f"OpenFisca-UK {self.variable_label} {self.statistic} absolute error is less than {self.max_error:,} in {self.year}"


class RelativeErrorLessThan(StatisticTest):
    def test(self):
        absolute_error = abs(self.openfisca_uk - self.official)
        relative_error = absolute_error / self.official
        return relative_error < self.max_error

    def describe(self):
        return f"OpenFisca-UK {self.variable_label} {self.statistic} relative error is less than {self.max_error:,} in {self.year}"


tests = []

for variable in statistics:
    variable_tests = statistics[variable]["test"]
    test_names = list(
        map(
            lambda item: item
            if isinstance(item, str)
            else list(item.keys())[0],
            variable_tests,
        )
    )
    dict_tests = {
        list(item.keys())[0]: item
        for item in variable_tests
        if isinstance(item, dict)
    }
    if "closer_than_ukmod" in test_names:
        for statistic in ("aggregate", "caseload"):
            if statistic in statistics[variable]["ukmod"]:
                for year in statistics[variable]["ukmod"][statistic]:
                    tests += [CloserThanUKMOD(variable, statistic, year)]
    if "aggregate_error_less_than" in test_names:
        for statistic in ("aggregate", "caseload"):
            if statistic in statistics[variable]["official"]:
                for year in statistics[variable]["official"][statistic]:
                    tests += [
                        AbsoluteErrorLessThan(
                            variable,
                            statistic,
                            year,
                            max_error=dict_tests["aggregate_error_less_than"][
                                "aggregate_error_less_than"
                            ],
                        )
                    ]
    if "relative_error_less_than" in test_names:
        for statistic in ("aggregate", "caseload"):
            if statistic in statistics[variable]["official"]:
                for year in statistics[variable]["official"][statistic]:
                    tests += [
                        RelativeErrorLessThan(
                            variable,
                            statistic,
                            year,
                            max_error=dict_tests["relative_error_less_than"][
                                "relative_error_less_than"
                            ],
                        )
                    ]


@pytest.mark.parametrize("test", tests, ids=str)
def test_statistics(test):
    assert test.test()



from openfisca_uk.model_api import *
import pandas as pd


class person_id(Variable):
    value_type = int
    entity = Person
    label = "ID for the person"
    definition_period = YEAR


class people(Variable):
    value_type = float
    entity = Person
    label = "Variable holding people"
    definition_period = YEAR
    default_value = 1


class raw_person_weight(Variable):
    value_type = float
    entity = Person
    label = "Weight factor"
    documentation = "Used to translate from survey respondents to UK residents"
    definition_period = YEAR
    default_value = 1


class person_weight(Variable):
    value_type = float
    entity = Person
    label = "Weight (region-adjusted)"
    documentation = "Adjusted to match regional population estimates"
    definition_period = YEAR

    def formula(person, period, parameters):
        nation = person.household("country", period)
        frs_person_weight = person("raw_person_weight", period)
        national_population = parameters(
            period
        ).demographic.population_estimate[nation]
        national_weight_sum = (
            pd.Series(frs_person_weight).groupby(nation).sum()[nation]
        )
        scale_factor = national_population / national_weight_sum
        return scale_factor * frs_person_weight


class age(Variable):
    value_type = int
    entity = Person
    label = "Age"
    unit = "year"
    documentation = "The age of the person in years"
    definition_period = YEAR
    quantity_type = STOCK

    def formula(person, period, parameters):
        ADULT_DEFAULT_AGE = 18
        CHILD_DEFAULT_AGE = 10
        is_adult = person.benunit.members_role == BenUnit.ADULT
        return where(is_adult, ADULT_DEFAULT_AGE, CHILD_DEFAULT_AGE)


class birth_year(Variable):
    value_type = int
    entity = Person
    label = "The birth year of the person"
    definition_period = YEAR

    def formula(person, period, parameters):
        return period.start.year - person("age", period)


class over_16(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is over 16"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) >= 16


class is_adult(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is an adult"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) >= 18


class is_child(Variable):
    value_type = bool
    entity = Person
    label = "Is a child"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) < 18


class child_index(Variable):
    value_type = int
    entity = Person
    label = "Child reference number"
    definition_period = YEAR

    def formula(person, period, parameters):
        # The child index, by age, descending (e.g. "first child" = 1)
        child_ranking = (
            person.get_rank(
                person.benunit,
                -person("age", period),
                condition=person("is_child", period),
            )
            + 1
        )
        # Fill in adult values
        return where(person("is_child", period), child_ranking, 100)


class is_eldest_child(Variable):
    label = "Is the eldest child"
    documentation = (
        "Whether this person is the eldest child in the benefit unit"
    )
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person("child_index", period) == 1


class is_benunit_eldest_child(Variable):
    value_type = bool
    entity = Person
    label = "Eldest child in the benefit unit"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        is_child = person("is_child", period)
        eldest_age = person.benunit("eldest_child_age", period)
        age_tie = person.benunit.sum((age == eldest_age) & is_child) > 1
        is_eldest_age = person("age", period) == eldest_age
        child_id = person("person_id", period) * is_child
        max_child_id = person.benunit.max(child_id)
        has_max_child_id = child_id == max_child_id
        return where(is_eldest_age & age_tie, has_max_child_id, is_eldest_age)


class MaritalStatus(Enum):
    SINGLE = "Single"
    MARRIED = "Married"
    SEPARATED = "Separated"
    DIVORCED = "Divorced"
    WIDOWED = "Widowed"


class marital_status(Variable):
    value_type = Enum
    possible_values = MaritalStatus
    default_value = MaritalStatus.SINGLE
    entity = Person
    label = "Marital status"
    definition_period = ETERNITY

    def formula(person, period, parameters):
        return where(
            person.benunit("is_married", period),
            MaritalStatus.MARRIED,
            MaritalStatus.SINGLE,
        )


class EducationType(Enum):
    NOT_IN_EDUCATION = "Not in education"
    PRE_PRIMARY = "Pre-primary"
    NOT_COMPLETED_PRIMARY = "Not completed primary"
    PRIMARY = "Primary"
    LOWER_SECONDARY = "Lower Secondary"
    UPPER_SECONDARY = "Upper Secondary"
    POST_SECONDARY = "Post Secondary"
    TERTIARY = "Tertiary"


class current_education(Variable):
    value_type = Enum
    possible_values = EducationType
    default_value = EducationType.NOT_IN_EDUCATION
    entity = Person
    label = "Current education"
    definition_period = YEAR


class highest_education(Variable):
    value_type = Enum
    possible_values = EducationType
    default_value = EducationType.UPPER_SECONDARY
    entity = Person
    label = "Highest status education completed"
    definition_period = YEAR


class in_FE(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is in Further Education"
    definition_period = YEAR
    set_input = set_input_dispatch_by_period


class in_HE(Variable):
    value_type = bool
    entity = Person
    label = "In higher education"
    definition_period = YEAR
    reference = "Whether this person is in Higher Education"
    set_input = set_input_dispatch_by_period


class Gender(Enum):
    MALE = "Male"
    FEMALE = "Female"


class gender(Variable):
    value_type = Enum
    possible_values = Gender
    default_value = Gender.MALE
    entity = Person
    label = "Gender of the person"
    definition_period = ETERNITY


class is_male(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is male"
    definition_period = ETERNITY

    def formula(person, period, parameters):
        return person("gender", period) == Gender.MALE


class is_female(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is female"
    definition_period = ETERNITY

    def formula(person, period, parameters):
        return person("gender", period) == Gender.FEMALE


class is_household_head(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is the head-of-household"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person.get_rank(person.household, person("age", period)) == 0


class is_benunit_head(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is the head-of-family"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person.get_rank(person.benunit, person("age", period)) == 0


class in_social_housing(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person lives in social housing"
    definition_period = YEAR

    def formula(person, period, parameters):
        tenure = person.household("tenure_type", period.this_year)
        tenures = tenure.possible_values
        return is_in(tenure, tenures.RENT_FROM_COUNCIL, tenures.RENT_FROM_HA)


class is_WA_adult(Variable):
    value_type = bool
    entity = Person
    label = "Whether is a working-age adult"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("is_adult", period) & ~person("is_SP_age", period)


class is_young_child(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is under 14"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period.this_year) < 14


class age_under_18(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is under age 18"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) < 18


class age_18_64(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is age 18 to 64"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return (age >= 18) & (age <= 64)


class age_over_64(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is over age 64"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) > 64


class is_older_child(Variable):
    value_type = bool
    entity = Person
    label = "Whether the person is over 14 but under 18"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return (age >= 14) & (age < 18)



from openfisca_uk.model_api import *


class main_residence_value(Variable):
    label = "Main residence value"
    documentation = "Total value of the main residence"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    quantity_type = STOCK


class other_residential_property_value(Variable):
    label = "Other residence value"
    documentation = (
        "Total value of all residential property owned by the household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    quantity_type = STOCK


class residential_property_value(Variable):
    label = "Residential property value"
    documentation = "Total value of all owned residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    quantity_type = STOCK

    def formula(household, period, parameters):
        return household("main_residence_value", period) + household(
            "other_residential_property_value", period
        )


class non_residential_property_value(Variable):
    label = "Non-residential property value"
    documentation = (
        "Total value of all non-residential property owned by the household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    quantity_type = STOCK


class property_wealth(Variable):
    label = "Property wealth"
    documentation = "Total property wealth across all owned properties"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    quantity_type = STOCK

    def formula(household, period):
        VARIABLES = [
            "residential_property_value",
            "non_residential_property_value",
        ]
        return add(household, period, VARIABLES)
